import .LinearAlgebra ;
import .LeastSquares ;
import .Random ;
import .Functions ;
import .Fourier ;


//! @decl array Higuchi( array serie, int kmin , int kmax, float error_percent , void|function stepper)
//! This function computes the Higuchi fractal dimension of a time serie.
//! @param serie
//! The time serie we want to analize.
//! @param kmin
//! Minimum value of the paramether we have to use to compute the fractal dimension
//! @param kmax
//! Maximum value of the paramether we have to use to compute the fractal dimension
//! @param error_percent
//! Percentual error of each point in serie
//! @param stepper
//! A function which, given a value of k, returns the next value of k to use.

array Higuchi( array serie, int kmin , int kmax, float error_percent , void|function stepper)
{
  float suma = 0.0 ;
  int n = sizeof(serie);
  int num_k = kmax - kmin + 1 ;
  array llista_k = ({}) ,  llista_lk = ({} ) ,  llista_errors = ({}) ;
  float sumak = 0.0 ;
  int index = -1 ;
  function f ;
  int k = kmin ;
  if ( stepper )
    {
      f = stepper ;
    }
  else
    {
      f = lambda ( int a ) { return a + 1 ; } ;
    }
  while ( k <= kmax )
    {
      sumak = 0.0 ;
      ++index ;
      for ( int m = 0 ; m < k ; ++m )
	{
	  int i_max = ( n - m - 1 ) / k ;
	  for ( int i = 0 ; i < i_max ; ++i )
	    {
	      float element_1 = serie[ (m + (i + 1) * k ) ] ;
	      float element_2 = serie[ (m + i*k) ] ;
	      suma += abs( element_1 - element_2 );
	    }
	  suma = suma * ( (float) ( n - 1 ) / ( i_max * k * k * k ) ) ;
	  sumak += suma ;
	}
      llista_k += ({log((float)k) }) ;
      llista_lk += ({log(sumak)});
      if (log(sumak) == 0 )
	{
	  llista_errors += ({ 0.1 }) ;
	}
      else
	{
	  llista_errors += ({ log(sumak) * error_percent }) ;
	}
      k = f(k);
    }
  
  array resultat = LinearLeastSquaresStraightLine ( ColumnFromArray (llista_k) , ColumnFromArray( llista_lk) , ColumnFromArray(llista_errors) ) ;
  return ({ - resultat[0] , resultat[1] , resultat[2] , resultat[3] , llista_k , llista_lk , llista_errors }) ;
}

//! Generates n points following a linear distribucion y = a x + b ;
array GenerateLinearData( float a, float b, int n ) 
{
  array serie = ( array ) NMatrix( n , 1 ) ;
  for ( int i = 0 ; i < n ; ++i )
    {
      serie[i] = a * i + b ;
    }
  return serie ;
}

//! Generates n points following a sinusoidal function. Only one period.
array GenerateSinusoidalData (int n ) 
{
  array serie = ( array ) NMatrix(n, 1 ) ;
  for ( int i = 0 ; i < n ; ++i )
    {
      serie[i] = cos ( 2 * Math.pi * i / n ) ;
    }
  return serie ;
}

//! Generates n points from the logistic map x -> r x (1-x)
array GenerateLogisticData( float r, float x0, int n , int transitori) 
{
  array serie = ( array ) NMatrix( n , 1 ) ;
  float valor = x0 ;
  if ( ! ( r < 4.0 && r > 0.0 ) )
    {
      throw(({"Invalid r value\n", backtrace()}));
    }
  for ( int i = 0 ; i < transitori ; ++i )
    {
       valor = valor * r * ( 1 - valor ) ;
    }
  for ( int i = 0 ; i < n ; ++i )
    {
      valor = valor * r * ( 1 - valor ) ;
      serie[i] = valor ;
      //write("Valor "  + i + " = " + valor + " \n");
    }
  return serie ;
}


//! Generates n points from a random walk (also called Fractional Brownian Function ), with fractal dimension 1.5 .  
array GenerateHiguchiData ( int n ) 
{
  int n_aleatoris = n + 1000 , limit_superior = 0 ;;
  array serie = ( array ) NMatrix(n,1) ;
  array aleatoris = ( array ) NMatrix ( n + 1000 , 1 ) ;
  float suma = 0.0 ;
  RandomSelfInit();
  for(int i = 0 ; i < n_aleatoris ; ++i )
    {
      aleatoris[i] = GaussDev( 0.0 , 1.0 ) ;
    }
  for( int i = 0 ; i < n ; ++i )
    {
      suma = 0.0 ;
      limit_superior = i + 1000 ;
      for ( int k = 0; k < limit_superior ; ++k )
	{
	  suma += aleatoris[k] ;
	}
      serie[i] = suma ;
    }
  return serie ;
}

//! @decl array BurlagaKlein ( array serie, int kmin , int kmax, float error_percent , void|function stepper)
//! This function computes the Burlaga and Klein's fractal dimension of a time serie.
//! @param serie
//! The time serie we want to analize.
//! @param kmin
//! Minimum value of the paramether we have to use to compute the fractal dimension
//! @param kmax
//! Maximum value of the paramether we have to use to compute the fractal dimension
//! @param error_percent
//! Percentual error of each point in serie
//! @param stepper
//! A function which, given a value of k, returns the next value of k to use.

array BurlagaKlein ( array serie, int kmin , int kmax, float error_percent , void|function stepper) 
{
  float suma1 = 0.0, suma2 = 0.0 ;
  int n = sizeof(serie);
  int num_k = kmax - kmin + 1 ;
  array llista_k = ({});
  array llista_lk = ({});
  array llista_errors = ({});
  float sumak = 0.0 ;
  int index = -1 ;
  int nmax = 0 ;
  float terme1, terme2;
  function f ;
  if ( stepper )
    {
      f = stepper ;
    }
  else
    {
      f = lambda ( int k ) { return k + 1 ; } ;
    }
  int k = kmin ;
  while ( k <= kmax )
    {
      nmax = (n - k)/k ;
      sumak = 0.0 ;
      ++index ;
      for(int i = 0 ; i < nmax ; ++i )
	{
	  suma1 = 0.0 ;
	  suma2 = 0.0 ;
	  for(int m = 0 ; m < k ; ++m )
	    {
	      terme1 = serie[m + i*k] ;
	      terme2 = serie[m + (i + 1 )*k];
	      suma1 += terme1 ;
	      suma2 += terme2 ;
	    }	
	  sumak += abs( suma1 - suma2 );
	}
      sumak = sumak / ((float) k*k) ;
      llista_k += ({log( (float) k) });
      llista_lk += ({log( sumak ) });
      if ( log(sumak) == 0.0 )
	{
	  llista_errors += ({0.1}) ;
	}
      else 
	{
	  llista_errors += ({log(sumak) * error_percent}) ;
	}
      k = f(k);
    }
  array resultat = LinearLeastSquaresStraightLine(ColumnFromArray(llista_k) , ColumnFromArray(llista_lk) , ColumnFromArray(llista_errors ));
  return ({ -resultat[0] , resultat[1] , resultat[2], resultat[3] , llista_k, llista_lk, llista_errors}) ;
}


//! @decl array SimpleFractalDimension( array serie, int kmin , int kmax, float error_percent, void|function stepper )
//! This function computes the fractal dimension of a time serie simply using as the curve length abs( y1 - y2 ). Very primitive method, It's here only to compare with BK and Higuchi methods.
//! @param serie
//! The time serie we want to analize.
//! @param kmin
//! Minimum value of the paramether we have to use to compute the fractal dimension
//! @param kmax
//! Maximum value of the paramether we have to use to compute the fractal dimension
//! @param error_percent
//! Percentual error of each point in serie
//! @param stepper
//! A function which, given a value of k, returns the next value of k to use.

array SimpleFractalDimension( array serie, int kmin , int kmax, float error_percent, void|function stepper )
{
  float suma = 0.0 ;
  int n = sizeof(serie);
  int num_k = kmax - kmin + 1 ;
  array llista_k = ({});
  array llista_lk = ({}) ;
  array llista_errors = ({}) ;
  int index = -1 ;
  int k = kmin ;
  function f ;
  if ( stepper )
    {
      f = stepper ;
    }
  else 
    {
      f = lambda ( int a ) { return a + 1 ; } ;
    }
  while( k <= kmax )
    {
      suma = 0.0 ;
      ++index ;
      int i_max = ( n  - 1 ) / k ;
      for ( int i = 0 ; i < i_max ; ++i )
	{
	  float element_1 = serie[  (i + 1) * k  ] ;
	  float element_2 = serie[  i*k ] ;
	  suma += abs( element_1 - element_2 );
	}
      suma = suma * ( (float) ( n - 1 ) / ( i_max * k * k  ) ) ;
      llista_k += ({ log((float)k) }) ;
      llista_lk += ({log(suma) });
      if (log(suma) == 0 )
	{
	  llista_errors += ({ 0.1 }) ;
	}
      else
	{
	  llista_errors +=  ({log(suma) * error_percent }) ;
	}
      k = f(k);
    }
  array resultat = LinearLeastSquaresStraightLine ( ColumnFromArray(llista_k) , ColumnFromArray( llista_lk) , ColumnFromArray(llista_errors) ) ;
  return ({ - resultat[0] , resultat[1] , resultat[2] , resultat[3], llista_k , llista_lk, llista_errors }) ;
}


//! Generates n points from the Weierstrass function. s is the fractal dimension of the graph. lamda must be greather than 1. 
array GenerateWeierstrassData ( int n , float xmin, float xmax , float lamda , float s )
{
  function f = Weierstrass( lamda, s  );
  array serie = ( array ) NMatrix( n , 1 ) ;
  float valor = xmin ;
  for ( int i = 0 ; i < n ; ++i )
    {
      valor += (xmax - xmin ) / n * i  ;
      
      serie[i] = f(valor) ;
      //write(serie[i] + "\n");
    }
  return serie ;
}


//! Generates points following an exponential power expectrum law, where P(k) ~ k^(-exponent). The angle is picked from a Uniform distribution between 0 and angle_max (in degree).
array GeneratePowerSpectrumExponentData (int n, float exponent, float angle_max )
{
  array ck = ({});
  array resultat = ({});
  float pi = atan(1.0) * 4.0 ;
  array teta = ({}) ;
  int N = (n / 2 ) * 2 ;
  for( int k = 1 ; k <= N / 2 ; ++k )
    {
      ck += ({ sqrt( n * pow( k , -exponent ) ) }); 
      teta += ({ UniformDeviation( 0.0 , angle_max ) });
    }
  float suma = 0.0 ;
  float angle = 0.0;
  for ( int j = 1 ; j <= N ; ++j )
    {
      suma = 0.0 ;
      for( int k = 1 ; k <= N / 2 ; ++k )
	{
	  angle = teta[k - 1] ;
	  suma += ck[k - 1] * cos( 2 * pi * j * k / N - angle ) ;
	}
      //write(suma + "\n" );
      resultat += ({ suma }) ;
    }
  return resultat ;
}


//! Tries to do a log-log adjust to get the exponent from a power spectrum which follows an exponential law.
array PowerSpectrumExponent( array serie, int kmax , float error_p)
{
  array aux = PowerSpectrum( serie, kmax );
  int N = sizeof(aux);
  array pspec = aux[1..(N - 1)] ;
  N = N - 1 ;
  array logk = ({});
  array logpower = ({});
  array llista_errors = ({}) ;
  float valor = 0.0 ;
  for ( int i = 0 ; i < N ; ++i )
    {
      logk += ({ log((float) i + 1 ) }) ;
      valor = log(pspec[i] );
      logpower += ({ valor }) ;
      if ( valor == 0.0 )
	{
	  llista_errors += ({ error_p }) ;
	}
      else
	{
	  llista_errors += ({ valor * error_p }) ;
	}
    }
  return LinearLeastSquaresStraightLine ( ColumnFromArray(logk) , ColumnFromArray( logpower) , ColumnFromArray(llista_errors) ) ;
}

int main ()
{
  array serie  = GenerateLinearData ( 1.0 , 2.5 , 100 ) ;
  array resultat ;
  function stepper = lambda ( int k ) { return k * 2 ; } ;
  write("Generant dades d'Higuchi per a la comparativa...\n");
  serie = GenerateHiguchiData( 10000 ) ;
  write("Ajustant les dades...\n");
  resultat = Higuchi( serie, 1 , 1000 , 0.1 , stepper );
  write("La dimensio fractal del test no 1 ha resultat ser " +  resultat[0] + "\n" );
  resultat = BurlagaKlein( serie, 1 , 1000 , 0.1 , stepper) ;
  write("El mateix calcul fet amb el metode de Burlaga i Klein dona com a resultat " +  resultat[0] + "\n" );
  resultat = SimpleFractalDimension( serie, 1 , 1000 , 0.1, stepper) ;
  write("El mateix calcul amb el metode simple dona " + resultat[0] + "\n");
  serie = GenerateSinusoidalData( 100 ) ;
  resultat = Higuchi( serie, 1, 20 , 0.1 ) ;
  write("La dimensio fractal del test no 2 ha resultat ser " + resultat[0] + "\n" );
  resultat = BurlagaKlein( serie, 1 , 20 , 0.1 ) ;
  write("El mateix calcul fet amb el metode de Burlaga i Klein dona com a resultat " +  resultat[0] + "\n" );
  resultat = SimpleFractalDimension( serie , 1 , 20 , 0.1) ;
  write("El mateix càlcul amb el metode simple dona " + resultat[0] + "\n");
  write("Generant dades per al test d'Higuchi...\n");
  serie = GenerateHiguchiData( 10000 ) ;
  write("Ajustant les dades generades... \n");
  resultat = Higuchi( serie, 1, 50, 0.1 ) ;
  write("La dimensio fractal del test no 3 ha resultat ser " + resultat[0] + "\n" );
  resultat = BurlagaKlein( serie, 1 , 50 , 0.1 ) ;
  write("El mateix calcul fet amb el metode de Burlaga i Klein dona com a resultat " +  resultat[0] + "\n" );
  resultat = SimpleFractalDimension( serie, 1 , 50 , 0.1) ;
  write("El mateix calcul amb el metode simple " + resultat[0] + "\n");
  write("Generant serie de dades extensiva\n");
  serie = GenerateLinearData( 2.5, 7.8 ,  50000 ) ;
  write("Calculant dimensio fractal\n");
  resultat = Higuchi ( serie , 1 , 50 , 0.1 ) ;
  write("El resultat del test extensiu en memoria ha sigut " + resultat[0] + "\n" );
  resultat = BurlagaKlein( serie, 1 , 50 , 0.1 ) ;
  write("El mateix calcul fet amb el metode de Burlaga i Klein dona com a resultat " +  resultat[0] + "\n" );
  resultat = SimpleFractalDimension( serie , 1 , 50 , 0.1 );
  write("El mateix amb el metode simple dona " + resultat[0] + "\n");
  write("Generant dades de Weirstrass...\n");
  serie = GenerateWeierstrassData( 5000 , 0.0 , 6.0 , 1.5 , 1.2 ) ;
  resultat = Higuchi(serie , 1 , 100 , 0.1 ); 
  serie = GeneratePowerSpectrumExponentData( 5000 , 2.0 , 360.0 );
  resultat = Higuchi(serie, 20 , 1000 , 0.01, stepper);
  write("La dimensio fractal de les potncies ha resultat ser " + resultat[0] + "\n" );
  serie = GenerateLogisticData( 3.98745 , 0.2 , 500, 1000 );
  resultat = Higuchi( serie , 1 , 50 , 0.1 );
  write("El resultat de les dades logístiques ha sigut " + resultat[0] + "\n" );

  serie = GeneratePowerSpectrumExponentData( 5000 , 2.0 , 360.0 ) ;
  resultat = PowerSpectrumExponent( serie, 20 , 0.01 );
  write("L'exponent de la serie de potencies es " + -resultat[0] + " +- " + resultat[2] + "\n" );
    return 0;
}


