//! @decl float Bisection ( function f, array range , float error_p, int iterations ) 
//! This function tries to find a root of f , given the interval (a,b) where you think that It's fitted. 
//! @param f 
//! The function that we are tryint to find a root.
//! @param range
//! The interval ({ a, b }) where we think that the root is fitted.
//! @param error_p 
//! The relative error that It's allowed to the solution.
//! @param iterations
//! The maxim number of iterations that the method can do trying to find the root.

float Bisection ( function f, array range , float error_p, int iterations ) 
{
  if ( f(range[0]) * f(range[1]) > 0 )
    {
      // Un error!!
      throw( ({"Non fitted root\n" , backtrace() }) ) ;
    }
  else
    {
      float a = range[0];
      float b = range[1];
      float fa = f(a);
      float fb = f(b);
      float root = (a + b) / 2.0 ;
      float fr = f(root);
      float tolerance = (abs (a - b) * error_p) ;
      float error_pas ;
      for ( int iter = 0 ; iter < iterations ; ++iter )
	{
	  root = (a + b) / 2.0 ;
	  fa = f(a);
	  fb = f(b);
	  fr = f(root);
	  if ( abs(fr) < tolerance ) //S'ha trobat una arrel
	    {
	      return root ;
	    }
	  else if ( fa * fr < 0 ) 
	    {
	      b = root ;
	    }
	  else if ( fb * fr < 0 )
	    {
	      a = root ;
	    }
	  error_pas = abs ( a - b );
	  if ( error_pas < tolerance ) //Hem trobat una arrel
	    {
	      return root ;
	    }
	}
      throw( ({ "Root not found\n" , backtrace() }));
    }
}

//! @decl float NewtonRaphson ( function f, function df, float point , float error_abs , int iterations ) 
//! This function tries to find a root of f using the Newton-Raphson method.
//! @param f
//! The function
//! @param df
//! The derivative of f
//! @param point
//! A point to start the method from x = point
//! @param error_abs 
//! The absolute error that we tolerate to the solution
//! @param iterations 
//! The maximum number of iterations that we can do trying to find the root.
float NewtonRaphson ( function f, function df, float point , float error_abs , int iterations ) 
{
  float root = point ;
  float df0 = df(root);
  float fr;

  for ( int iter = 0 ; iter < iterations ; ++iter )
    {
      df0 = df(root);
      if ( df0 == 0 )
	{
	  //Error! Divisio per zero!
	  write("Divisio per zero.\n");
	}
      fr = f(root);
      root = root - fr / df0 ;
      fr = f(root);
      if ( abs(fr) < error_abs )
	{
	  return root ;
	}
    }
  throw( ({ "Root not found\n" , backtrace() }));
}

//! @decl float Steffensen ( function f, float point , float error_abs , int iterations )
//! This function tries to find a root of f using the Newton-Raphson method.
//! @param f
//! The function
//! @param point
//! A point to start the method from x = point
//! @param error_abs 
//! The absolute error that we tolerate to the solution
//! @param iterations 
float Steffensen ( function f, float point , float error_abs , int iterations )
{
  float root = point ;
  float df0 ;
  float fr ;
  for ( int iter = 0 ; iter < iterations ; ++iter )
    {
      df0 = f(root + f(root)) - f(root) ;
      if ( df0 == 0 )
	{
	  //Error!
	  write("Error de divisió per zero\n");
	}
      else
	{
	  fr = f(root);
	  root = root - fr*fr/df0 ;
	  fr = f(root);
	  if ( abs(fr ) < error_abs )
	    {
	      return root ;
	    }
	}
    }
  throw( ({ "Root not found\n" , backtrace() }));
}


//! @decl float RegulaFalsi ( function f, array range , float error_p, int iterations ) 
//! This function tries to find a root of f , given the interval (a,b) where you think that It's fitted. It uses the Regula-Falsi method. The paramethers have the same meaning than in Bisection. 
float RegulaFalsi ( function f, array range , float error_p, int iterations ) 
{
  if ( f(range[0]) * f(range[1]) > 0 )
    {
      // Un error!!
      throw( ({"Non fitted root\n" , backtrace() }) ) ;
    }
  else
    {
      float a = range[0];
      float b = range[1];
      float fa = f(a);
      float fb = f(b);
      float root = (a + b) / 2.0 ;
      float fr = f(root);
      float tolerance = (abs (a - b) * error_p) ;
      float error_pas ;
      for ( int iter = 0 ; iter < iterations ; ++iter )
	{
	  root = ( a * fb - b * fa) / ( fb - fa ) ;
	  fa = f(a);
	  fb = f(b);
	  fr = f(root);
	  if ( abs(fr) < tolerance ) //S'ha trobat una arrel
	    {
	      return root ;
	    }
	  else if ( fa * fr < 0 ) 
	    {
	      b = root ;
	    }
	  else if ( fb * fr < 0 )
	    {
	      a = root ;
	    }
	  error_pas = abs ( a - b );
	  if ( error_pas < tolerance ) //Hem trobat una arrel
	    {
	      return root ;
	    }
	}
      throw( ({ "Root not found\n" , backtrace() }));
    }
}


int main ()
{
  function f = sin ;
  function f2 = cos ;
  float arrel = Bisection(f, ({ -0.72 , 2.12 }) , 0.0001 , 100 );
  write("S'ha trobat l'arrel " + arrel + "\n" );
  arrel = NewtonRaphson(f, f2, 3.12, 0.01 , 100 );
  write("S'ha trobat l'arrel " + arrel + "\n" );
  arrel = Steffensen ( f , 3.12 , 0.01 , 100 ) ;
  write("S'ha trobat l'arrel " + arrel + "\n");
  arrel = RegulaFalsi( f, ({ -0.72 , 2.12 }) , 0.0001 , 100 ) ;
  write("S'ha trobat l'arrel " + arrel + "\n");
  return 0;
}


